#pragma once

#ifndef EXCEPTION_H
#define EXCEPTION_H


/**
	\file
	\brief Contains Exception class.
*/




#include <string>
#include <stdexcept>




/*==============================================================================
	THROW_EXCEPTION macros
==============================================================================*/


/// \brief Throws an Exception labeled with the file, line and string \p what.
#define THROW_EXCEPTION( what ) \
	throw ::Exception( __FILE__, __LINE__, ( what ) )

/// \brief Throws an Exception labeled with the file, line, and strings \p what and \p message. Typically, \p what will describe the failure, and \p message will be a library error message.
#define THROW_EXCEPTION_MESSAGE( what, message ) \
	throw ::Exception( __FILE__, __LINE__, ( what ), ( message ) )




/*==============================================================================
	Exception class
==============================================================================*/


/**
	\brief Default exception class.

	Exceptions should not be constructed explicitly. Instead, use the
	THROW_EXCEPTION() and THROW_EXCEPTION_MESSAGE() macros, which will
	construct and throw exceptions annotated with the filename and line number
	of the failure.

	The error message returned by what() has the following components:

	1. Filename and line number at which the exception was thrown.
	2. Description string \p what passed to the constructor.
	3. (Optional) string \p message passed to the constructor, intended to be used to store error messages from external libraries (although the WindowsException and OpenCL::Exception classes handle this for \c HRESULT error codes and OpenCL error codes, respectively).
*/
struct Exception : public std::exception {

	Exception( char const* const file, unsigned int const line, std::string const& what );
	Exception( char const* const file, unsigned int const line, std::string const& what, char const* const message );


	virtual ~Exception() throw();


	virtual const char* what() const throw();    ///< Returns text (UTF-8) error message for this exception.


private:

	std::string m_what;
};




#endif    // EXCEPTION_H
