#pragma once

#ifndef OPENCL_DELETER_H
#define OPENCL_DELETER_H


/**
	\file
	\brief Contains OpenCL::Deleter class.
*/




#include "settings.h"




#if ( SETTING_OPENCL != 0 )




#include "opencl_exception.h"

#include <CL/cl.h>




namespace OpenCL {




/*==============================================================================
	Deleter class
==============================================================================*/


/**
	\brief A deleter class for OpenCL objects, for use in stl smart pointers.

	\param t_Type  OpenCL pointee type.

	Specializations of this class contain a single method, <tt>operator()</tt>,
	which passes the given <tt>t_Type*</tt> through to the appropriate OpenCL
	function. For example, for memory objects, <tt?t_Type = _cl_mem</tt> (the
	more familiar <tt>cl_mem</tt> is simply <tt>_cl_mem*</tt>), and
	<tt>Deleter< _cl_mem >::operator()</tt> will call
	<tt>clReleaseMemObject</tt>.
*/
template< typename t_Type >
struct Deleter;




/*==============================================================================
	Deleter specializations
==============================================================================*/


/**
	\brief Deleter specialization for OpenCL contexts.

	\see \ref Deleter
*/
template<>
struct Deleter< _cl_context > {

	inline void operator()( _cl_context* const pointer ) {

		if ( pointer != NULL )
			CHECK_OPENCL( clReleaseContext( pointer ) );
	}
};


/**
	\brief Deleter specialization for OpenCL command queues.

	\see \ref Deleter
*/
template<>
struct Deleter< _cl_command_queue > {

	inline void operator()( _cl_command_queue* const pointer ) {

		if ( pointer != NULL )
			CHECK_OPENCL( clReleaseCommandQueue( pointer ) );
	}
};


/**
	\brief Deleter specialization for OpenCL memory objects.

	\see \ref Deleter
*/
template<>
struct Deleter< _cl_mem > {

	inline void operator()( _cl_mem* const pointer ) {

		if ( pointer != NULL )
			CHECK_OPENCL( clReleaseMemObject( pointer ) );
	}
};


/**
	\brief Deleter specialization for OpenCL samplers.

	\see \ref Deleter
*/
template<>
struct Deleter< _cl_sampler > {

	inline void operator()( _cl_sampler* const pointer ) {

		if ( pointer != NULL )
			CHECK_OPENCL( clReleaseSampler( pointer ) );
	}
};


/**
	\brief Deleter specialization for OpenCL programs.

	\see \ref Deleter
*/
template<>
struct Deleter< _cl_program > {

	inline void operator()( _cl_program* const pointer ) {

		if ( pointer != NULL )
			CHECK_OPENCL( clReleaseProgram( pointer ) );
	}
};


/**
	\brief Deleter specialization for OpenCL kernels.

	\see \ref Deleter
*/
template<>
struct Deleter< _cl_kernel > {

	inline void operator()( _cl_kernel* const pointer ) {

		if ( pointer != NULL )
			CHECK_OPENCL( clReleaseKernel( pointer ) );
	}
};


/**
	\brief Deleter specialization for OpenCL events.

	\see \ref Deleter
*/
template<>
struct Deleter< _cl_event > {

	inline void operator()( _cl_event* const pointer ) {

		if ( pointer != NULL )
			CHECK_OPENCL( clReleaseEvent( pointer ) );
	}
};




}    // namespace OpenCL




#endif    // SETTING_OPENCL




#endif    // OPENCL_DELETER_H
