#include "octree.h"
#include "synchronized.h"
#include "renderer.h"
#include "vector.h"
#include "exception.h"
#include "helpers.h"
#include "settings.h"

#ifdef _OPENMP
#include <omp.h>
#endif    // _OPENMP

#include <SDL/SDL.h>
#include <SDL/SDL_ttf.h>

#include <iostream>
#include <array>
#include <memory>
#include <cmath>
#include <stdexcept>

#include <stdlib.h>
#include <time.h>
#include <sys/time.h>




/*==============================================================================
	Global variables
==============================================================================*/


double g_fps = 0;
double g_rps = 0;




/*==============================================================================
	Redraw function
==============================================================================*/


template< int t_ColorFormat, int t_Platform, int t_RendererPlatform >
void Redraw(
	SDL_Surface* pSurface,
	TTF_Font* pFont,
	Synchronized::Heap< t_Platform >& heap,
	Synchronized::TextureHeap< t_ColorFormat, t_Platform >& textureHeap,
	typename Synchronized::Heap< t_Platform >::template Pointer< Octree::Node< t_ColorFormat, t_Platform > > const& pRoot,
	Renderer::State< t_RendererPlatform > const& renderer,
	Renderer::Camera const& camera
)
{
	if ( SDL_FillRect( pSurface, NULL, SDL_MapRGB( pSurface->format, 0, 0, 0 ) ) != 0 )
		THROW_EXCEPTION_MESSAGE( "SDL_FillRect failed!", SDL_GetError() );

	if ( SDL_LockSurface( pSurface ) != 0 )
		THROW_EXCEPTION_MESSAGE( "SDL_LockSurface failed!", SDL_GetError() );
	try {    // **TODO: RAII?

		Vector< unsigned int, 2 > const dimensions = { { pSurface->w, pSurface->h } };

#if 0
		typename Synchronized::Heap< t_Platform >::template Pointer< Octree::Node< t_ColorFormat, t_Platform > > const pSubtractedRoot = Octree::Subtract(
			heap,
			textureHeap,
			pRoot,
			g_coordinates,
			0.1,
			2
		);
		heap.Synchronize();
		textureHeap.Synchronize();
#else    // 0/1
		typename Synchronized::Heap< t_Platform >::template Pointer< Octree::Node< t_ColorFormat, t_Platform > > const& pSubtractedRoot = pRoot;
#endif    // 0/1

		renderer.TraceScreen(
			static_cast< uint8_t* >( pSurface->pixels ),
			dimensions,
			pSurface->pitch,
			heap,
			textureHeap,
			pSubtractedRoot,
			camera
		);
	}
	catch( ... ) {

		SDL_UnlockSurface( pSurface );
		throw;
	}
	SDL_UnlockSurface( pSurface );

	{	char strings[ 4 ][ 128 ];

#ifdef _OPENMP
		unsigned int const threads = omp_get_max_threads();
#else    // _OPENMP
		unsigned int const threads = 1;
#endif    // _OPENMP
		snprintf( strings[ 0 ], ARRAYLENGTH( strings[ 0 ] ), "threads = %u, resolution = %dx%d", threads, pSurface->w, pSurface->h );
		snprintf( strings[ 1 ], ARRAYLENGTH( strings[ 1 ] ), "%.2f Mrays / frame", pSurface->w * pSurface->h / 1000000.0 );
		snprintf( strings[ 2 ], ARRAYLENGTH( strings[ 2 ] ), "%.2f frames / sec", g_fps );
		snprintf( strings[ 3 ], ARRAYLENGTH( strings[ 3 ] ), "%.2f Mrays / sec", g_rps / 1000000 );

		SDL_Color const textColor = { 0xff, 0xff, 0xff };

		Sint16 xx = 8;
		Sint16 yy = 8;
		for ( unsigned int ii = 0; ii < ARRAYLENGTH( strings ); ++ii ) {

			strings[ ii ][ ARRAYLENGTH( strings[ ii ] ) - 1 ] = '\0';

			std::unique_ptr< SDL_Surface, void (*)( SDL_Surface* ) > pTextSurface( TTF_RenderText_Blended( pFont, strings[ ii ], textColor ), SDL_FreeSurface );
			if ( pTextSurface.get() == NULL )
				THROW_EXCEPTION_MESSAGE( "TTF_RenderText_Blended failed!", TTF_GetError() );

			SDL_Rect destinationRect = { xx, yy, static_cast< Uint16 >( pTextSurface->w ), static_cast< Uint16 >( pTextSurface->h ) };
			if ( SDL_BlitSurface( pTextSurface.get(), NULL, pSurface, &destinationRect ) != 0 )
				THROW_EXCEPTION_MESSAGE( "SDL_BlitSurface failed!", SDL_GetError() );
			yy += pTextSurface->h;
		}
	}
}




/*==============================================================================
	main function
==============================================================================*/


int main( int argc, char* argv[] ) {

	int result = EXIT_SUCCESS;

	try {

		// ==== initialize voxel and keyboard data ====

		//enum { COLOR_FORMAT = COLOR_FORMAT_RGB_565 };
		enum { COLOR_FORMAT = COLOR_FORMAT_RGBA_8888 };
#if ( SETTING_OPENCL != 0 )
		std::shared_ptr< OpenCL::State > pState( new OpenCL::State() );
		Synchronized::Heap< PLATFORM_OPENCL > heap( pState, ( 1u << 25 ) );
		Synchronized::TextureHeap< COLOR_FORMAT, PLATFORM_OPENCL > textureHeap( pState, 27 );
		Synchronized::Heap< PLATFORM_OPENCL >::Pointer< Octree::Node< COLOR_FORMAT, PLATFORM_OPENCL > > pRoot = Octree::Create( heap, textureHeap );
		Renderer::State< PLATFORM_OPENCL > renderer( pState );
#else    // SETTING_OPENCL
		Synchronized::Heap< PLATFORM_HOST > heap;
		Synchronized::TextureHeap< COLOR_FORMAT, PLATFORM_HOST > textureHeap;
		Synchronized::Heap< PLATFORM_HOST >::Pointer< Octree::Node< COLOR_FORMAT, PLATFORM_HOST > > pRoot = Octree::Create( heap, textureHeap );
		Renderer::State< PLATFORM_HOST > renderer;
#endif    // SETTING_OPENCL
		heap.Synchronize();
		textureHeap.Synchronize();

		Vector< double, 3 > const initialX = { { 1, 0, 0 } };
		Vector< double, 3 > const initialY = { { 0, 1, 0 } };
		Vector< double, 3 > const initialZ = { { 0, 0, 1 } };
		Vector< double, 3 > const initialPosition = { { 0.5, 0.5, 1.5 } };
		Renderer::Camera camera( initialX, initialY, initialZ, initialPosition );

		enum MovementKeys {

			MOVEMENT_KEY_MOVE_FORWARD = 0,
			MOVEMENT_KEY_MOVE_BACK,
			MOVEMENT_KEY_MOVE_DOWN,
			MOVEMENT_KEY_MOVE_UP,
			MOVEMENT_KEY_MOVE_RIGHT,
			MOVEMENT_KEY_MOVE_LEFT,
			MOVEMENT_KEY_TURN_COUNTERCLOCKWISE,
			MOVEMENT_KEY_TURN_CLOCKWISE,
			MOVEMENT_KEY_TURN_RIGHT,
			MOVEMENT_KEY_TURN_LEFT,
			MOVEMENT_KEY_TURN_UP,
			MOVEMENT_KEY_TURN_DOWN,
			MOVEMENT_KEY_ZOOM_IN,
			MOVEMENT_KEY_ZOOM_OUT,

			MOVEMENT_KEYS_SIZE
		};
		bool movementKeys[ MOVEMENT_KEYS_SIZE ];
		std::fill( movementKeys, movementKeys + ARRAYLENGTH( movementKeys ), false );

		double const fpsRate = 0.95;
		double const rateTurn = 0.8;
		double const rateZoom = 0.4;
		double const rateMove = 0.2;

		// ==== initialize SDL ====

		if ( SDL_Init( SDL_INIT_VIDEO ) != 0 )
			THROW_EXCEPTION_MESSAGE( "SDL_Init failed!", SDL_GetError() );
		atexit( SDL_Quit );

		std::unique_ptr< SDL_Surface, void (*)( SDL_Surface* ) > pSurface( SDL_SetVideoMode( 800, 600, 32, SDL_HWSURFACE | SDL_DOUBLEBUF ), SDL_FreeSurface );
		if ( pSurface.get() == NULL )
			THROW_EXCEPTION_MESSAGE( "SDL_SetVideoMode failed!", SDL_GetError() );

		SDL_WM_SetCaption( "Voxel", "Voxel" );

		// ==== initialize SDL font support ====

		if ( TTF_Init() != 0 )
			THROW_EXCEPTION_MESSAGE( "TTF_Init failed!", TTF_GetError() );
		atexit( TTF_Quit );

		std::unique_ptr< TTF_Font, void (*)( TTF_Font* ) > pFont( TTF_OpenFont( "verdanab.ttf", 14 ), TTF_CloseFont );
		if ( pFont.get() == NULL )
			THROW_EXCEPTION_MESSAGE( "TTF_OpenFont failed!", TTF_GetError() );

		// ==== message loop ====

		double frames  = 0;
		double rays    = 0;
		double seconds = 0;
		double ticks = 0;
		{	struct timeval time;
			gettimeofday( &time, NULL );
			ticks = time.tv_sec + static_cast< double >( time.tv_usec ) / 1000000;
		}

		SDL_Event event;
		for ( bool done = false; ! done; ) {

			// ==== draw frame ====

			Redraw( pSurface.get(), pFont.get(), heap, textureHeap, pRoot, renderer, camera );
			SDL_Flip( pSurface.get() );

			// ==== update counters ====

			double const ticksOld = ticks;
			{	struct timeval time;
				gettimeofday( &time, NULL );
				ticks = time.tv_sec + static_cast< double >( time.tv_usec ) / 1000000;
			}
			double const deltaSeconds = ticks - ticksOld;

			frames = fpsRate * frames + ( 1 - fpsRate );
			rays = fpsRate * rays + ( 1 - fpsRate ) * pSurface->w * pSurface->h;
			seconds = fpsRate * seconds + ( 1 - fpsRate ) * deltaSeconds;

			g_fps = frames / seconds;
			g_rps = rays   / seconds;

			// ==== process messages ====

			while ( SDL_PollEvent( &event ) ) {

				switch( event.type ) {

					case SDL_KEYUP: {

						switch( event.key.keysym.sym ) {
							case SDLK_w:        { movementKeys[ MOVEMENT_KEY_MOVE_FORWARD          ] = false; break; }
							case SDLK_s:        { movementKeys[ MOVEMENT_KEY_MOVE_BACK             ] = false; break; }
							case SDLK_PAGEDOWN: { movementKeys[ MOVEMENT_KEY_MOVE_DOWN             ] = false; break; }
							case SDLK_PAGEUP:   { movementKeys[ MOVEMENT_KEY_MOVE_UP               ] = false; break; }
							case SDLK_d:        { movementKeys[ MOVEMENT_KEY_MOVE_RIGHT            ] = false; break; }
							case SDLK_a:        { movementKeys[ MOVEMENT_KEY_MOVE_LEFT             ] = false; break; }
							case SDLK_q:        { movementKeys[ MOVEMENT_KEY_TURN_COUNTERCLOCKWISE ] = false; break; }
							case SDLK_e:        { movementKeys[ MOVEMENT_KEY_TURN_CLOCKWISE        ] = false; break; }
							case SDLK_RIGHT:    { movementKeys[ MOVEMENT_KEY_TURN_RIGHT            ] = false; break; }
							case SDLK_LEFT:     { movementKeys[ MOVEMENT_KEY_TURN_LEFT             ] = false; break; }
							case SDLK_UP:       { movementKeys[ MOVEMENT_KEY_TURN_UP               ] = false; break; }
							case SDLK_DOWN:     { movementKeys[ MOVEMENT_KEY_TURN_DOWN             ] = false; break; }
							case SDLK_HOME:     { movementKeys[ MOVEMENT_KEY_ZOOM_OUT              ] = false; break; }
							case SDLK_END:      { movementKeys[ MOVEMENT_KEY_ZOOM_IN               ] = false; break; }
							case SDLK_ESCAPE: { done = true; break; }
							default: break;
						}
						break;
					}

					case SDL_KEYDOWN: {

						switch( event.key.keysym.sym ) {
							case SDLK_w:        { movementKeys[ MOVEMENT_KEY_MOVE_FORWARD          ] = true; break; }
							case SDLK_s:        { movementKeys[ MOVEMENT_KEY_MOVE_BACK             ] = true; break; }
							case SDLK_PAGEDOWN: { movementKeys[ MOVEMENT_KEY_MOVE_DOWN             ] = true; break; }
							case SDLK_PAGEUP:   { movementKeys[ MOVEMENT_KEY_MOVE_UP               ] = true; break; }
							case SDLK_d:        { movementKeys[ MOVEMENT_KEY_MOVE_RIGHT            ] = true; break; }
							case SDLK_a:        { movementKeys[ MOVEMENT_KEY_MOVE_LEFT             ] = true; break; }
							case SDLK_q:        { movementKeys[ MOVEMENT_KEY_TURN_COUNTERCLOCKWISE ] = true; break; }
							case SDLK_e:        { movementKeys[ MOVEMENT_KEY_TURN_CLOCKWISE        ] = true; break; }
							case SDLK_RIGHT:    { movementKeys[ MOVEMENT_KEY_TURN_RIGHT            ] = true; break; }
							case SDLK_LEFT:     { movementKeys[ MOVEMENT_KEY_TURN_LEFT             ] = true; break; }
							case SDLK_UP:       { movementKeys[ MOVEMENT_KEY_TURN_UP               ] = true; break; }
							case SDLK_DOWN:     { movementKeys[ MOVEMENT_KEY_TURN_DOWN             ] = true; break; }
							case SDLK_HOME:     { movementKeys[ MOVEMENT_KEY_ZOOM_OUT              ] = true; break; }
							case SDLK_END:      { movementKeys[ MOVEMENT_KEY_ZOOM_IN               ] = true; break; }
							default: break;
						}
						break;
					}

					case SDL_QUIT: {

						done = true;
						break;
					}
				}
			}

			// ==== update position ====

			{	double const move = ( movementKeys[ MOVEMENT_KEY_MOVE_FORWARD ] ? 1.0 : 0.0 ) - ( movementKeys[ MOVEMENT_KEY_MOVE_BACK ] ? 1.0 : 0.0 );
				if ( move != 0 ) {

					double const rate = move * rateMove * deltaSeconds;
					camera.MoveForward( rate );
				}
			}
			{	double const move = ( movementKeys[ MOVEMENT_KEY_MOVE_DOWN ] ? 1.0 : 0.0 ) - ( movementKeys[ MOVEMENT_KEY_MOVE_UP ] ? 1.0 : 0.0 );
				if ( move != 0 ) {

					double const rate = move * rateMove * deltaSeconds;
					camera.MoveDown( rate );
				}
			}
			{	double const move = ( movementKeys[ MOVEMENT_KEY_MOVE_RIGHT ] ? 1.0 : 0.0 ) - ( movementKeys[ MOVEMENT_KEY_MOVE_LEFT ] ? 1.0 : 0.0 );
				if ( move != 0 ) {

					double const rate = move * rateMove * deltaSeconds;
					camera.MoveRight( rate );
				}
			}
			{	double const turn = ( movementKeys[ MOVEMENT_KEY_TURN_CLOCKWISE ] ? 1.0 : 0.0 ) - ( movementKeys[ MOVEMENT_KEY_TURN_COUNTERCLOCKWISE ] ? 1.0 : 0.0 );
				if ( turn != 0 ) {

					double const rate = turn * rateTurn * deltaSeconds;
					camera.TurnClockwise( rate );
				}
			}
			{	double const turn = ( movementKeys[ MOVEMENT_KEY_TURN_RIGHT ] ? 1.0 : 0.0 ) - ( movementKeys[ MOVEMENT_KEY_TURN_LEFT ] ? 1.0 : 0.0 );
				if ( turn != 0 ) {

					double const rate = turn * rateTurn * deltaSeconds;
					camera.TurnRight( rate );
				}
			}
			{	double const turn = ( movementKeys[ MOVEMENT_KEY_TURN_DOWN ] ? 1.0 : 0.0 ) - ( movementKeys[ MOVEMENT_KEY_TURN_UP ] ? 1.0 : 0.0 );
				if ( turn != 0 ) {

					double const rate = turn * rateTurn * deltaSeconds;
					camera.TurnDown( rate );
				}
			}
			{	double const zoom = ( movementKeys[ MOVEMENT_KEY_ZOOM_OUT ] ? 1.0 : 0.0 ) - ( movementKeys[ MOVEMENT_KEY_ZOOM_IN ] ? 1.0 : 0.0 );
				if ( zoom != 0 ) {

					double const rate = zoom * rateZoom * deltaSeconds;
					camera.ZoomOut( rate );
				}
			}
		}
	}
	catch( std::exception& exception ) {

		std::cerr << "caught exception: " << exception.what() << std::endl;
		result = EXIT_FAILURE;
	}
	catch( ... ) {

		std::cerr << "caught exception: <unknown error>" << std::endl;
		result = EXIT_FAILURE;
	}

	return result;
}
